# Copyright 2023 Facundo Batista
# Licensed under the GPL v3 License
# For further info, check https://github.com/facundobatista/pyempaq

"""Expose the version information at module level, as is standard."""

from pyempaq._version import __version__, VERSION  # noqa
